---
# Render runtime args
test_stdout_arg: "{{ test_stdout_arg }}"
test_stdout_arg2: "{{ test_stdout_arg2 }}"
test_runtime_arg: "{{ test_runtime_arg }}"

# Render cachefile args
test_string_args: "{{ test_string_args }}"
test_integer_args: "{{ test_integer_args }}"
test_float_args: "{{ test_float_args }}"
test_boolean_args: "{{ test_boolean_args }}"
test_nested_args: {{ test_nested_args }}
test_array_args: {{ test_array_args }}

# Render inline complex args
test_runtime_complex_arg_list: {{ test_runtime_complex_arg_list }}
test_runtime_complex_arg_int: {{ test_runtime_complex_arg_int }}
test_runtime_complex_arg_dict: {{ test_runtime_complex_arg_dict }}
test_runtime_complex_arg_float: {{ test_runtime_complex_arg_float }}
